// Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
// Built in San Francisco.

// This software is distributed under a commercial license for personal,
// educational, corporate or any other use.
// The software as a whole or any parts of it is prohibited for distribution or
// use without obtaining a license from 3D Control Systems, Inc.

// All software licenses are subject to the 3DPrinterOS terms of use
// (available at https://www.3dprinteros.com/terms-and-conditions/),
// and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import QtQuick 2.7
import QtQuick.Controls 2.2
import Qt.labs.folderlistmodel 2.13

import AppStyle 1.0
import AppControls 1.0

import "constants.js" as Constants


Item {
    id: selectFileView
    objectName: Constants.States.LOCAL

    function formatBytes(bytes, decimals) {
        if(bytes === 0) return '0 Bytes';
        var k = 1024,
        dm = decimals <= 0 ? 0 : decimals || 2,
        sizes = ['Bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'],
        i = Math.floor(Math.log(bytes) / Math.log(k));
        return parseFloat((bytes / Math.pow(k, i)).toFixed(dm)) + ' ' + sizes[i];
    }

    FolderListModel {
        id: folderListModel
        rootFolder: "/media"
        nameFilters: ["*.gcode"]
        showDirsFirst: true
    }

    Item {
        id: listItem
        anchors {
            top: parent.top
            left: parent.left
            right: parent.right
            bottom: footer.top
            bottomMargin: 10
        }

        ListView {
            id: fileListView
            model: folderListModel
            clip: true
            anchors.fill: parent
            currentIndex: -1

            property bool showHead: false
            property bool scrollBarEnabled: fileListView.contentHeight > fileListView.height
            property real defaultItemHeight: 64
            property int subFoldersCount: 0

            ScrollBar.vertical: ScrollBar {
                id: scroll
                policy: ScrollBar.AlwaysOn
                visible: fileListView.scrollBarEnabled
                parent: fileListView.parent
                anchors {
                    top: parent.top
                    left: parent.right
                    bottom: parent.bottom
                }

                contentItem: Rectangle {
                    implicitWidth: 7
                    implicitHeight: 100
                    color: scroll.pressed ? AppStyle.textColor : AppStyle.foreground
                }
            }

            header: Rectangle {
                id: headDotDot
                width: fileListView.width
                height: visible ? fileListView.defaultItemHeight : 0
                color: headDotDotMouseArea.pressed ? AppStyle.foreground : "transparent"
                visible: fileListView.subFoldersCount > 0
                border {
                    color: AppStyle.listBorderColor
                    width: 1
                }

                Text {
                    id: dotdotText
                    anchors {
                        left: parent.left
                        leftMargin: 30
                    }
                    height: fileListView.defaultItemHeight
                    font.pixelSize: fileListView.defaultItemHeight * 0.45
                    text: ".."
                    color: AppStyle.textColor
                    horizontalAlignment: Text.AlignLeft
                    verticalAlignment: Text.AlignVCenter
                    elide: Text.ElideRight
                }

                MouseArea {
                    id: headDotDotMouseArea
                    anchors.fill: parent
                    onClicked: {
                        var folderStr = folderListModel.folder.toString()
                        folderListModel.folder = folderStr.split('/').slice( 0, -1 ).join('/');
                        fileListView.subFoldersCount--
                        fileListView.currentIndex = -1

                        if (folderListModel.folder.toString().split("//")[1] === folderListModel.rootFolder.toString()) {
                            console.log("return to parent dir")

                        }
                    }
                }
            }

            delegate: Rectangle {
                id: fileListDelegate
                width: fileListView.width
                height: fileListView.defaultItemHeight
                color: (fileListDelegateMouseArea.pressed || fileListDelegate.ListView.isCurrentItem) ? AppStyle.foreground : "transparent"

                border {
                    color: AppStyle.listBorderColor
                    width: 1
                }

                Row {
                    id: fileListRow
                    height: fileListView.defaultItemHeight
                    anchors {
                        top: parent.top
                        left: parent.left
                        right: parent.right
                        rightMargin: 30
                        leftMargin: 30
                        verticalCenter: parent.verticalCenter
                    }

                    Text {
                        id: filenameText
                        height: fileListView.defaultItemHeight
                        width: fileListRow.width * 0.75
                        font.pixelSize: fileListView.defaultItemHeight * 0.35
                        text: fileName
                        color: AppStyle.textColor
                        horizontalAlignment: Text.AlignLeft
                        verticalAlignment: Text.AlignVCenter
                        elide: Text.ElideRight
                    }

                    Text {
                        id: sizeText
                        width: fileListRow.width * 0.25
                        height: fileListView.defaultItemHeight
                        font.pixelSize: fileListView.defaultItemHeight * 0.35
                        text: fileIsDir ? qsTr("Folder") : selectFileView.formatBytes(fileSize)
                        color: AppStyle.textColor
                        horizontalAlignment: Text.AlignHCenter
                        verticalAlignment: Text.AlignVCenter
                    }
                }

                MouseArea {
                    id: fileListDelegateMouseArea
                    anchors.fill: parent
                    onClicked: {
                        if (fileIsDir) {
                            folderListModel.folder += "/" + fileName
                            fileListView.subFoldersCount++
                            fileListView.currentIndex = -1
                            return
                        }

                        fileListView.currentIndex = index
                    }
                }

                Behavior on height { PropertyAnimation { duration: 250 } }
            }
        }
    }

    Footer {
        id: footer
        anchors {
            bottom: parent.bottom
            right: parent.right
            left: parent.left
        }

        nextBtnText: qsTr("Print")
        nextBtnVisible: fileListView.currentIndex !== -1

        onBackClicked: {
            rootItem.clearHeadMsg()
            rootItem.back()

        }

        onNextClicked: {
            var filePath = folderListModel.get(fileListView.currentIndex, "filePath")
            console.log("sending file to print: " + filePath)
            filesModel.printFile(filePath)
        }
    }

    Component.onCompleted: {
        rootItem.setHeadMsg(qsTr("SD card"))
    }
}
